/*--------------------------------------------------------------------------*/
/* What is it?
    OneCopy.Exe is a program that will transfer files from one floppy to
    another, using memory buffers instead of having to copy files to a RAM
    disk (with possible space limitations) or to your hard drive (and having
    to deal with this extra directory on your HD).

    Originally this question came up for a BBS user who has an AT with no
    hard drive, but a 5-1/4 drive and a 3-1/2 drive. Making a copy of a file
    on one single disk format was a bit tricky.

    Personally, it was just a learning experience. This is the first
    program I've dealt with that used FAR calls, or extensive memory
    buffers. The BEST way to learn programming is simply to get in there
    and practice, this is practice.  If you find this program useful, I'd
    like to here about it. If you find problems, or suggestions for
    improvements, I'd like to hear about those too. There are no
    guaruntees with a program like this, if it breaks in two, you own both
    halves. But it might help you learn a little about programming or
    provide you with some useful ideas.

    Copyright 1988, Doug Boone. P.O. Box 5108, Chico, CA. 95928-5108
	Data Phone (916) 893-9019 24 hours.  Voice (916) 891-0748

	10-09-88
		Added qsort procedure to sort files by size and start at the
		biggest. This _should_ give a little extra efficiency when
		copying a disk with more files than memory can handle.

	10-11-88
		Made the number of files that onecopy can handle larger (120).
		If you need still _more_ files, then you'll need to increase the
		value in brackets where it says:
		struct  afile  list[120];

		Compile in the COMPACT memory model.

	11-28-88
		Fixed bug in copying large files. Note for today: You have to
		manually advance a pointer when reading really large files into
		memory buffers! Thanks Kevin Silcox.
*/
/*--------------------------------------------------------------------------*/

#include    <stdio.h>
#include    <string.h>
#include    <io.h>
#include    <dir.h>
#include    <fcntl.h>
#include    <sys\stat.h>
#include    <alloc.h>
#include    <errno.h>

#define     GOT_FILE    0x01
#define     COPIED_FILE 0x02

#ifndef __COMPACT__
#error ONECOPY Must be compiled in the COMPACT memory model!
#endif

struct	one_file {
    unsigned    long    size;
    char    name[13];
    unsigned    char    done;
    unsigned    char    *buff_ptr;
    };

	struct      one_file    list[120];
    int         highest;        /* number of files looked at */
    char        drive;
    char        file_path[80];
    unsigned    long        free_mem;

    int     cdecl   get_matches(char *);
    int     cdecl   read_files(int);
    int     cdecl   write_files(int);
    void    cdecl   handle_big(int);
    unsigned    long    cdecl write_big(int,unsigned long,unsigned long);
    unsigned    long    cdecl read_big(int,unsigned long);
    int     cdecl    compare(unsigned long *,unsigned long *);


main(int argc,char    *argv[])
{

    char    *result;
    int     matches = 0;        /* Number of files matching command line */
    int     red = 0;            /* Number of files read in */
    int     done = 0;            /* Number of files written out */
    int        i;
    char    ch;

        strcpy(file_path,argv[1]);
        printf("\n\nOneCopy, Single-Drive file copy program by Doug Boone Copyright 1988,\n");
        printf("For FREE use and distribution ONLY! (FidoNet 119/5)\n\n");

        if (argc < 2) {
            printf("\n\nUse as: ONECOPY A:\path\*.* or ONECOPY a:command.com\n");
            printf("OneCopy will copy one or more files, and will expand * or ?\n");
            printf("wildcards.\n\n");
            exit(1);
            }

/* Seperate the file path and file name */
        result = strrchr(file_path,'\\');
        if (result != NULL) {
            result++;
            *result=0x00;
            }
        else {
            result = strrchr(file_path,':');
            if (result !=NULL) {
                result++;
                *result = 0x00;
                }
            else file_path[0] = 0x00;
            }

/* Get the drive letter being used  */
		result = strrchr(file_path,':');
        if (result != NULL) {
            drive = toupper(file_path[0]);
            if (strlen(file_path) > 2)
                drive = toupper(file_path[0]);
            }
        else
            drive = getdisk()+ '@';
        if (drive > 'B') {
            printf("\n WARNING!! You're attempting to copy a file on drive %c onto itself.\n");
            printf("           If this is a hard-drive you could get a SERIOUS error!\n\n");
            printf("Continue? [Y,n] ");
            ch = toupper(getch());
            if (ch == 'N')
                exit(0);
                }
/* Let the fun begin... */
        printf("\nInsert SOURCE disk in drive %c:, press key when ready....\n",
            drive);
        if ((ch =getch()) == 0x03)
            exit(5);

/* Look for files that match the command line arguement */
        if ((matches = get_matches(argv[1])) == 0) {
            printf("No matching file(s) found for %s\n\n",argv[1]);
            exit(1);
            }

		qsort(&list[0].size,matches,sizeof(list[0]),compare);

/* Find out how much memory is available */
    free_mem = farcoreleft() - 5120L;
/* If there are any files that are bigger than free memory, */
/* special handling will be necessary. */
	for (i=0;i<matches;i++) {
        if (list[i].size > free_mem)
            handle_big(i);
            }
/* For files that are smaller than free memory */
    do {
        if (red>0) {
            printf("\nInsert SOURCE disk in drive %c:, press key when ready....\n",
                drive);
        if ((ch =getch()) == 0x03)
            exit(5);
            }
        red = read_files(matches);
        done = write_files(matches);
		} while (done < matches);
/* All done */
        exit(0);
}

int compare(unsigned long *size1, unsigned long *size2)
{
    if (*size2 < *size1)
        return(-1);
    else return(*size2 > *size1);
}


/* Look for files that match the command line. Expand wildcards. */
int get_matches(char    *match)
{

    struct    ffblk    ffblk;
    int        done;
    int     i=0;

    done = findfirst(match,&ffblk,0);
	while ((done == 0)&& (i<120)) {
        strcpy(list[i].name,ffblk.ff_name);
        list[i].size = ffblk.ff_fsize;
        list[i].done = 0;
        i++;
        done = findnext(&ffblk);
        }
	return(i);
}

/* Start grabbing files from the source disk */
int     read_files(int total)
{
    char    temp[64];
    int     i;
    int     handle;        /* file hanle for open() */
	unsigned    int    count;          /* block size for read() */
	unsigned    int    success;        /* success shold = count if read was good */
	unsigned    long   bytes_read;     /* number of bytes read in */
    int     files_read = 0;
    char    ch;

	for (i=0;i<total;i++){
#ifdef    DEBUG
        printf("On file %2i of %2i  %s\n",i,total,list[i].name);
#endif
        if (list[i].done == 0) {
            if ((list[i].buff_ptr = farmalloc(list[i].size)) != NULL) {
                sprintf(temp,"%s%s",file_path,list[i].name);
                if ((handle = open(temp,O_RDONLY|O_BINARY)) != NULL) {
					bytes_read = 0L;
                    printf("Reading %s\n",temp);
/* When using the C read() function the maximum buffer is 64k  */
/* but OneCopy can use bigger memory buffers so we'll need to  */
/* read and write in smaller blocks, count stores the size of  */
/* block to use                                                */
                    if (list[i].size > (unsigned long) 64512)
                        count = 64512;
                    else
                        count = (unsigned) list[i].size;
                    do {
						success = read(handle,(list[i].buff_ptr+bytes_read),count);
                        if (success != count) {
                            perror(temp);
                            exit(1);
                            }
						bytes_read += (unsigned long) count;
						if ((list[i].size - bytes_read) < count)
							count = (unsigned int) list[i].size - bytes_read;
						} while (bytes_read < list[i].size);
                    close(handle);
                    }         /* Done reading one file handle */
                list[i].done = GOT_FILE;
                }        /* Done with one buffer */
            }
        if ( list[i].done > 0 )
            files_read++;
        }        /* done _trying_ to read files */
    return(files_read);
}




int     write_files(int high)
{
    int     i;
    int     handle;
    unsigned    int     count;        /* size of output buffer */
    unsigned    int     success;        /* check for valid write() */
	unsigned    long    bytes_written;
    char    temp[80];
    int     files_written = 0;

    printf("\nInsert TARGET disk in drive %c, press any key...\n",drive);
	if ((getch()) == 0x03)
        exit(5);

    for(i=0;i<=high;i++) {
        if (list[i].done & GOT_FILE) {

            sprintf(temp,"%s%s",file_path,list[i].name);
            printf("Writing out %s\n",temp);
            handle = open(temp,O_BINARY|O_WRONLY|O_CREAT,S_IREAD|S_IWRITE);
            if (handle < 0) {
                perror("Error opening output file ");
                exit(1);
                }
            if (list[i].size > (unsigned long) 64512)
                count = 64512;
            else
                count = list[i].size;
			 bytes_written = 0;
                do {
					success = write(handle,(list[i].buff_ptr + bytes_written),count);
                    if (success != count) {
                        sprintf(temp,"Error writing %s ",list[i].name);
                        perror(temp);
                        close(handle);
                        exit(1);
                        }
					bytes_written += (unsigned long) count;
					if (((unsigned long) count) > (list[i].size - bytes_written))
						count = (unsigned) (list[i].size - bytes_written);
					} while (count > 0);
                farfree(list[i].buff_ptr);
                close(handle);
                list[i].done <<= 1;
                }        /* Done writing out one file */
            if (list[i].done & COPIED_FILE)
                files_written++;
            }        /* done writing out files */
    return(files_written);
}

/* Dealing with files that are bigger than the free memory buffer */
/* was pretty difficult. */

void    handle_big(int which)
{
    unsigned    long    bytes_read;
    unsigned    long    bytes_written;
    unsigned    long    total_done = 0L;

    printf("%s at %lu is larger than free memory %lu\n",
        list[which].name,list[which].size,free_mem);
    do {
        bytes_read = read_big(which,total_done);
        bytes_written = write_big(which,bytes_read,total_done);
        total_done += bytes_written;
        } while (total_done < list[which].size);
    list[which].done = COPIED_FILE;
    return;
}

unsigned    long    read_big(int  which, unsigned long done)
{
    unsigned    int        count = 64512;
    unsigned     int        success;
    unsigned    long    to_go;
	unsigned    long    bytes_read = 0L;
	unsigned	long	bytes_pass = 0L;
    static  char    temp[80];
    int     handle;
    int     i=10;

    sprintf(temp,"%s%s",file_path,list[which].name);

    to_go = list[which].size - done;

    if (to_go > free_mem)
        to_go = free_mem;

    do {
        list[which].buff_ptr = farmalloc(to_go);
        if (list[which].buff_ptr == NULL) {
            to_go = (to_go/10) * i;
            i--;
            }
        } while((list[which].buff_ptr == NULL) && i>0);
    if (list[which].buff_ptr == NULL) {
        printf("FarMalloc Error! %lu\n",free_mem);
        exit(1);
        }

    if (count > to_go)
        count = (unsigned int) to_go;

    handle = open(temp,O_RDONLY|O_BINARY);
    if (handle <0) {
        perror(temp);
        close(handle);
        exit(1);
        }
    lseek(handle,done,SEEK_SET);        /* Move to where we are in the file */
    do {
		success = read(handle,(list[which].buff_ptr + bytes_pass),count);
        if (success != count) {
            sprintf(temp,"Error reading %s %u %u",list[which].name,count,success);
            close(handle);
            exit(1);
            }
		bytes_read += (unsigned long) success;
		bytes_pass += (unsigned long) success;
        if ((to_go - bytes_read) < ((unsigned long) count))
            count = (unsigned int) (to_go - bytes_read);
        } while (bytes_read < to_go);
    close(handle);
    return(bytes_read);
}

/*--------------------------------------------------------------------------*/
unsigned long    write_big(int  which, unsigned long bytes_read, unsigned long  done)
{
    unsigned    int        count = 64512;
    unsigned    int        success;
    unsigned    long    to_go;
	unsigned    long    bytes_written=0L;
	unsigned	long	bytes_pass = 0L;
    static  char    temp[80];
    int     handle;

    printf ("Insert TARGET disk in drive %c, press any key.....\n",drive);
	if ((getch()) == 0x03)
        exit(5);

    sprintf(temp,"%s%s",file_path,list[which].name);
    to_go = bytes_read;


    if (count > to_go)
        count = (unsigned int) to_go;
    if (done>bytes_read)
        handle = open(temp,O_RDWR|O_BINARY|O_APPEND);
    else
        handle = open(temp,O_RDWR|O_BINARY|O_CREAT,S_IWRITE|S_IREAD);

    if (handle <0) {
        perror(temp);
        close(handle);
        exit(1);
        }

    do {
		success = write(handle,(list[which].buff_ptr + bytes_pass),count);
        if (success != count) {
            sprintf(temp,"Error writing %s %u %u",list[which].name,count,success);
            close(handle);
            exit(1);
            }
		bytes_written += (unsigned long) success;
		bytes_pass += (unsigned long) success;
        if ((to_go - bytes_written) < ((unsigned long) count))
            count = (unsigned int) (to_go - bytes_written);
        } while (bytes_written < to_go);
    close(handle);
    farfree(list[which].buff_ptr);

    printf("Insert SOURCE disk in drive %c:, press any key.....\n",drive);
	if ((getch()) == 0x03)
        exit(5);

    return(bytes_written);
}

